#include <assert.h>
#include <stdlib.h>
#include <stdio.h>
#include <ctype.h>
#include <string.h>
#include <math.h>
#include "config.h"

#define MAXTEXTL 128
#define MAXPATH  512
#define MAXFIT 1024

float *lam, *erlam, *sig, *ersig, *amp, *eramp, *flux, *errfl, *sbg;
int *ssel;

char    ion[MAXFIT];
char    stage[MAXFIT];
char   *iarr[MAXFIT];
char   *starr[MAXFIT];

char title1[MAXTEXTL];
char title2[MAXTEXTL];
char tabtit[MAXTEXTL];
char rname[MAXTEXTL];
int lpp=40,ndim=0,nar=0,np=0,i=0;

int aefflag=0;
float dt=1.;
float *aeff, *alam, *aval;

/* float *apos, *aneg; */

/*  -------------------------------------------------------------------- */

int read_log (char *path)
{
  FILE * fd;
  char line[BUFSIZ], *test1,*test;
  int items = 0, count_line=0,ilen=0,stlen=0,j=0;

  if (NULL == (fd = fopen(path, "r")))
    {
      printf("* Warning * %s: File not found.\n", path);
      return (-1);
    }

  while (fgets(line, BUFSIZ-1, fd))
    if (line[0] != '#') ++ndim;

  rewind(fd);

  if (ndim > MAXFIT) 
    {
      printf("* Warning * : Too many lines in file %s; increase MAXFIT in cora_tex.c and compile with make\n",path);
      return (-1);
    }

  lam    = (float *) malloc (sizeof(float) * ndim);
  erlam  = (float *) malloc (sizeof(float) * ndim);
  sig    = (float *) malloc (sizeof(float) * ndim);
  ersig  = (float *) malloc (sizeof(float) * ndim);
  amp    = (float *) malloc (sizeof(float) * ndim);
  eramp  = (float *) malloc (sizeof(float) * ndim);
  flux   = (float *) malloc (sizeof(float) * ndim);
  errfl   = (float *) malloc (sizeof(float) * ndim);
  sbg    = (float *) malloc (sizeof(float) * ndim);
  aval   = (float *) malloc (sizeof(float) * ndim);

  for (i=0;i<MAXFIT;++i) 
    {iarr[i]    = (char *) malloc (3);
     starr[i]    = (char *) malloc (19);}

  test   = malloc(19);
  test1   = malloc(1);

  if (lam==NULL || erlam==NULL || sig==NULL || ersig==NULL || amp==NULL || eramp==NULL || flux==NULL || errfl==NULL || sbg==NULL)
    {
      fprintf(stderr,"* Warning *: Out of memory\n");
      return(-1);
    }

  fgets(title1, MAXTEXTL, fd);

  while (fgets(line, BUFSIZ-1, fd))
    {
      if (line[0] != '#')
        {
          items = sscanf(line, "%f %f %f %f %f %f %f %s %s",
                    &lam[count_line], &erlam[count_line], &sig[count_line],
		    &ersig[count_line], &amp[count_line], &eramp[count_line],
		    &sbg[count_line], ion, stage);

	  flux[count_line]=amp[count_line]/dt;
	  errfl[count_line]=eramp[count_line]/dt;

	  stlen=strlen(stage);
	  ilen=strlen(ion);

          strncpy(iarr[count_line],"           ",4);
          strncpy(starr[count_line],"              ",7);

	  strncpy(iarr[count_line],ion,ilen);
	  strncpy(starr[count_line],stage,stlen);

	sprintf(test,"%c",tolower(starr[count_line][0]));
	for (j=1;j<stlen;++j)
	{
	  sprintf(test1,"%c",tolower(starr[count_line][j]));
	  strcat(test,test1);
	}

	strncpy(starr[count_line],test,stlen);

	sprintf(test,"%c",toupper(iarr[count_line][0]));
	for (j=1;j<ilen;++j)
	{
	  sprintf(test1,"%c",tolower(iarr[count_line][j]));
          strcat(test,test1);
	}

	strncpy(iarr[count_line],test,ilen);

          ++count_line;
        }
     }

  if (count_line < ndim)
    {
      printf("* Warning * Race condition detected\n");
      free(lam); free(erlam);free(sig);free(ersig);free(amp);free(eramp);free(flux);free(errfl);free(sbg);free(ion);
      return (-1);
    }

  return (ndim);
}

/*  -------------------------------------------------------------------- */
 
void
make_asciifile (char *path)
{
  FILE *fd;
  char su[MAXTEXTL], au[MAXTEXTL];
 
if (title1[27] == 'A')
  sprintf(su,"%s","\\AA");
else
  sprintf(su,"%s","km/s");
if (aefflag == 1)
  if (dt == 1.) sprintf(au,"%s","cts/cm^2"); else sprintf(au,"%s","cts/cm^2/ksec");
else
  if (dt == 1.) sprintf(au,"%s","cts"); else sprintf(au,"%s","cts/ksec");
 
sprintf(title1,"%s%s%s%s%s%s%s\n","lam[A] +- err  Sig[",su,"] +- err  A[cts] +- err  flux[",au,"] +-err  sbg[",au,"/A] ");

 
  fd = fopen (path, "w");

  fprintf(fd,"%s",title1);
  for (i=0;i<ndim;++i)
     fprintf(fd,"%f  %f %f %f %e %e %e %e %e %f\n"
           , lam[i], erlam[i], sig[i], ersig[i], amp[i], eramp[i], flux[i], errfl[i], sbg[i], aval[i]);
fclose(fd);
return;
 
} 

/*  -------------------------------------------------------------------- */

void
make_texfile (char *path)
{
  FILE *fd, *fd2;

  int j=0,id=0,iup,flag=0,k;
  char su[MAXTEXTL], au[MAXTEXTL];
  char dummy[2*MAXTEXTL];

  iup=lpp;

if (title1[27] == 'A')
  sprintf(su,"%s","\\AA");
else
  sprintf(su,"%s","km/s");
if (aefflag == 1)
  if (dt == 1.) sprintf(au,"%s","cts/cm$^2$"); else sprintf(au,"%s","cts/cm$^2$/ksec");
else
  if (dt == 1.) sprintf(au,"%s","cts"); else sprintf(au,"%s","cts/ksec");

sprintf(title1,"%s\n"," $\\lambda$  &   A & flux & A$_{\\rm eff}$\\\\");
sprintf(title2,"%s%s%s\n"," [\\AA]  &   [cts] & [",au,"] & [cm$^2$]\\\\");


      fd = fopen (path, "r");
      if(!fd)
        flag=1;
      else
        {
	  while(fgets(dummy,2.*MAXTEXTL,fd)) ++j;
	  rewind(fd);
	  fd2 = fopen ("172563811.tmp","w");
	  for (i=0;i<j-1;++i)  /* remove last line \end{document}  */
	    {
		fgets(dummy,2*MAXTEXTL,fd);
		fprintf(fd2,"%s",dummy);
	    }
	  fclose(fd2);
          fclose(fd);
	  sprintf(dummy,"rm %s",path);
	  system(dummy);
	  sprintf(dummy,"mv 172563811.tmp %s",path);
	  system(dummy);
        }

  fd = fopen (path, "a");

  if (flag==1)
   {
    fprintf (fd, "%s\n","\\documentstyle{article}");
    fprintf (fd, "%s\n","\\pagestyle{empty}");
    fprintf (fd, "%s\n","\\textwidth17cm");
    fprintf (fd, "%s\n","\\textheight26cm");
    fprintf (fd, "%s\n","\\voffset -2cm");
    fprintf (fd, "%s\n","\\hoffset -3cm");
    fprintf (fd, "%s\n","\\begin{document}");
    printf("Writing to %s...\n",path);
   }
  else printf("Appending to %s ...\n",path);

  for (j=0;j<np;++j)
   {
printf("Page %d of %d\n",j,np);
     fprintf (fd, "%s\n","\\begin{table*}");
     fprintf (fd, "%s\n","\\begin{flushleft}");
     fprintf (fd, "%s\n","\\renewcommand{\\arraystretch}{1.2}");
     sprintf(rname," %f to %f \\AA.",lam[id],lam[iup-1]);
     fprintf (fd, "%s%s%s\n","\\caption{",tabtit,"}");
     fprintf (fd, "{\\scriptsize\n");
     fprintf (fd, "%s\n","\\begin{tabular}{cccc}");
     fprintf (fd, "%s\n", title1);
     fprintf (fd, "%s\n", title2);

  for (i=id;i<iup;++i)
   {
     k=ssel[i];
     fprintf(fd,"%5.3f $\\pm$ %5.3f& %6.3f $\\pm$ %6.2f& %6.3f $\\pm$ %6.3f& %5.3f \\\\\n"
           , lam[k], erlam[k], amp[k], eramp[k], flux[k], errfl[k], aval[k]);
   }
     fprintf (fd, "%s\n","\\hline");
     fprintf (fd, "%s\n","\\end{tabular}");
     fprintf (fd, "}\n");
     fprintf (fd, "%s\n","\\renewcommand{\\arraystretch}{1}");
     fprintf (fd, "%s\n","\\end{flushleft}");
     fprintf (fd, "{\\tiny created with Cor{\\sc A} program \\today}\n");
     fprintf (fd, "%s\n","\\end{table*}");
     id=(j+1)*lpp;
     iup=(j+1)*lpp;
     if (ndim-i < lpp) lpp=ndim-i;
     iup += lpp;
   }
  fprintf (fd, "%s\n","\n\n\\end{document}");
  printf(".. tex file %s created.\n",path);

fclose(fd);
return;

}
 
/*  -------------------------------------------------------------------- */

int get_aeff(char *arfile)
{

FILE *fd;
int items=0,iflag=0;
char line[BUFSIZ];
float al,ar;
 
    if (NULL== (fd=fopen(arfile,"r")))
    {
      printf("* Warning * %s: File not found.\n", arfile);
      return(-1);
    }
    else {
      while (fgets(line, BUFSIZ-1, fd)) if (line[0] != '#') ++nar;
      rewind(fd);}

   aeff = (float *) malloc (sizeof(float) * nar); 
   alam = (float *) malloc (sizeof(float) * nar);
/*   apos = (float *) malloc (sizeof(float) * nar);
     aneg = (float *) malloc (sizeof(float) * nar); */

  if (alam==NULL || aeff==NULL /*|| apos==NULL || aneg==NULL*/)
    {
      fprintf(stderr,"* Warning *: Out of memory\n");
      return(-1);
    }

   i=0;
   while (fgets(line, BUFSIZ-1, fd))
    {
	al = 0.; ar = 0.;
	if (line[0] != '#')
	{items = sscanf(line, "%f %f %f", &alam[i], &al,&ar);
	aeff[i]=al+ar;  /* two columns can be given, to be added */
	if (aeff[i] > al && iflag == 0) {printf("You have given two columns with effective areas. They will be added.\n");++iflag;}
/*	apos[i]=al;
	aneg[i]=ar; */
/*  if (alam[i] < 20.) fprintf(stderr,"%d %f  %f  %f %f\n",i,alam[i],al,ar,aeff[i]); */
	++i;}
    }
return(0);
}

/*  -------------------------------------------------------------------- */ 

int main(int argc, char *argv[])
{

FILE * fd;
int j=0,k,isort=0;
char datafile[MAXPATH];
char texfile[MAXPATH];
char asciifile[MAXPATH];
char usage[MAXPATH];
char arfile[MAXPATH];
char sname[MAXTEXTL];

strncpy (datafile, "cora.dat", MAXPATH);
strncpy (texfile, "cora.tex", MAXPATH);
strncpy (asciifile, "cora.ascii", MAXPATH);
strncpy (usage, "Usage: cora_tex [-help] [-s] [-i logfile] [-l lpp] [-t outfile] [-a arfile] [-e expos. time] [-c caption]\n", MAXPATH);
strncpy (arfile, "cora.aef", MAXPATH);
strncpy (tabtit, "Fit results", MAXTEXTL);
strncpy (sname, "Logfile", MAXTEXTL);

  while (argc > 1 && argv[1][0] == '-')
    {
      switch (argv[1][1])
        {
         case 'l':
          ++argv;
          --argc;
          if (argc == 1)
             {
                fprintf (stderr,usage);
                exit (8);
             }
          lpp = atoi(argv[1]);
          break;
        case 'i':
          ++argv;
          --argc;
          if (argc == 1)
            {
              fprintf (stderr, usage);
              exit (8);
            }
          strncpy (datafile, argv[1], MAXPATH);
          break;
        case 'a':
          ++argv;
          --argc;
          if (argc == 1)
            {
              fprintf (stderr, usage);
              exit (8);
            }
          strncpy (arfile, argv[1], MAXPATH);
	  aefflag = 1;
          break;
	case 's':
	  isort=1;
	  break;
        case 'e':
          ++argv;
          --argc;
          if (argc == 1)
            {
              fprintf (stderr, usage);
              exit (8);
            }
	  dt = atof(argv[1])/1000.;
          break;
        case 't':
          ++argv;
          --argc;
          if (argc == 1)
            {
              fprintf (stderr, usage);
              exit (8);
            }
          strncpy (texfile, argv[1], MAXPATH);
          break;
        case 'c':
          ++argv;
          --argc;
          if (argc == 1)
            {
              fprintf (stderr, usage);
              exit (8);
            }
          strncpy (sname, argv[1], MAXTEXTL);
          break;
        default:
          fprintf (stderr, usage);
	  fprintf (stderr, "%s"," -s        :    sort by wavelengths\n");
          fprintf (stderr, "%s","logfile    : file with fit results\n");
          fprintf (stderr, "%s","lpp        : number of lines per page\n");
          fprintf (stderr, "%s","outfile    : TeX file (appended if present)\n");
          fprintf (stderr, "%s","arfile     : file with effective areas (two columns)\n");
          fprintf (stderr, "%s","expos time : Exposure time\n");
          fprintf (stderr, "%s","caption    : Caption for LeTeX table\n");
	  exit(8);
          break;
        }
      ++argv;
      --argc;
    }

  /*  read log file into global variables
   */
    if (NULL== (fd=fopen(datafile,"r"))) 
    {
      printf("* Warning * %s: File not found.\n", datafile);
      return(-1);
    } 
    else fclose(fd);

    ndim = read_log(datafile);
    assert (ndim !=-1);

    sprintf(tabtit,"%s (dt=%5.3f\\,ksec)",sname,dt);

    np=floor(1.*ndim/lpp)+1;

    if (ndim < lpp) lpp=ndim;

    if (aefflag == 1)
     {
       i=get_aeff(arfile);
       for (i=0;i<ndim;++i)
	 {
	   j=1;
	   while (lam[i] > alam[j] && j<nar) ++j;
	   aval[i] = aeff[j-1]+(aeff[j]-aeff[j-1])*(alam[j]-lam[i]);
	   flux[i]= flux[i] / aval[i];
	   errfl[i] = errfl[i] / aval[i];
	   sbg[i] = sbg[i];
	 }
      }

   ssel = (int *) malloc (sizeof(int) * ndim);
  if (isort==0) for (i=0;i<ndim;++i) ssel[i]=i; else for (i=0;i<ndim;++i) ssel[i]=0;

  if (isort > 0)
    for (i=0;i<ndim;++i)
      {
	k=0;
	for(j=0;j<ndim;++j) if (lam[i]>lam[j] &&i !=j ) ++k;
	ssel[k]=i;
      }

/* for (i=0;i<ndim;++i) fprintf(stderr,"%f ",lam[i]);
fprintf(stderr,"\n");
for (i=0;i<ndim;++i) fprintf(stderr,"%d ",ssel[i]);
fprintf(stderr,"\n");
for (i=0;i<ndim;++i) fprintf(stderr,"%f ",lam[ssel[i]]);
fprintf(stderr,"\n"); */

    make_texfile(texfile);

    make_asciifile(asciifile);

return (0);                                                                     
}
