#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <string.h>

#include "cora_fit.h"
#include "cpgplot.h"

#define MOUSE_R 'X'
#define MOUSE_M 'D'
#define MOUSE_L 'A'

#ifndef MAX
#define MAX(a,b) ((a) > (b) ? (a) : (b))
#endif
#ifndef MIN
#define MIN(a,b) ((a) < (b) ? (a) : (b))
#endif

/* -- for smoothing/rebinning --
 */
#define S_AVERAGE 0
#define S_MEDIAN  1
#define S_REBIN   2
static int smooth_mode = S_AVERAGE;
static int smooth      = 1;
static float zoom      = 1.0;

/*  ------------------------ */

static 
float * copy_to_float (Vector * inVec)
{
  int i;

  float * outVec = (float *) malloc (sizeof (float) * inVec->len);

  if (outVec == NULL)
    {
      fprintf (stderr, "* Warning *: copy_to_float: Out of memory\n");
      return NULL;
    }
  
  for (i = 0; i < inVec->len; ++i)
    outVec[i] = inVec->vec[i];

  return outVec;
}

/*  ------------------------ */

void redraw_it (float * xposition, float yposition, int lines, 
                float * wav, float * flux, int n_count,
                float uxmin, float uxmax, float uymin, float uymax)
{
  int i;
  double dx,dy;
  char smfac[64];
  
  cpgend();
  if(cpgopen("/XSERVE") != 1)
    {
      printf("* Warning *: Cannot open plot device\n");
      return;
    }
  
  
  cpgenv(uxmin, uxmax, uymin, uymax, 0, 1);
  cpglab("\\gl [\\A]", "Counts/bin", "");
  
  cpgbin(n_count, wav, flux, 1);
  cpgpt (n_count, wav, flux, 1);
  
  cpgsci(3);
  for (i = 0; i < lines; ++i)
    {
      cpgmove(xposition[i], uymin);
      cpgdraw(xposition[i], uymax);
    }
  cpgsci(1);
  if (yposition != 0.)
    {
      cpgsci(2);
      cpgmove(uxmin, yposition);
      cpgdraw(uxmax, yposition);
      cpgsci(1);
    }
  dx = uxmax - uxmin;
  dy = uymax - uymin;

  cpgtext(uxmin, uymax+.10*dy, "Left  Mouse Button: Select range");
  cpgtext(uxmin, uymax+.07*dy, "Right Mouse Button: Select line");
  cpgtext(uxmin, uymax+.04*dy, "d:    Undo selection of last line");
  cpgtext(uxmin, uymax+.01*dy, "b:    Select Background");
  
  cpgtext(uxmin+.65*dx, uymax+.10*dy, "q:     quit        c: restore");
  cpgtext(uxmin+.65*dx, uymax+.07*dy, "space: print cursor position");
  cpgtext(uxmin+.65*dx, uymax+.04*dy, "s:     cycle smoothing mode");
  cpgtext(uxmin+.65*dx, uymax+.01*dy, "+:     smooth    -: unsmooth");

  cpgtext(uxmin+.40*dx, uymax+.04*dy, "./,: slide up/down");
  cpgtext(uxmin+.40*dx, uymax+.07*dy, "</>: zoom out/in");

  if (smooth > 1)
    {
      if (smooth_mode == S_AVERAGE)
	sprintf(smfac, "average %d", smooth);
      if (smooth_mode == S_MEDIAN)
	sprintf(smfac, "median %d", smooth);
      if (smooth_mode == S_REBIN)
	sprintf(smfac, "rebin %d", smooth);
      cpgtext(uxmin+.45*dx, uymax+.01*dy, smfac);
    }

  return;
}

static float s_mod (float a, float b)
{
  float mod;
  if (b != 0) 
    mod = a-(int)(a/b)*b;
  else 
    return (-1);
  return (mod);
}

static int s_comp_func (const void *a, const void *b)
{
  float *x, *y;
  x = (float *) a;
  y = (float *) b;

  if (*x < *y)
      return -1;
  else if (*x > *y)
      return  1;
  else
      return 0;
}

float s_median(float *list, int num)

{
  float median;
  static float mm = 2.0;

  if (num == 0 || list == NULL)
    {
      fprintf (stderr, "* Warning *: s_median(): Invalid input.");
      return (0.0);
    }

  qsort(list, num, sizeof(float), s_comp_func);

  /* Calculate median 
   */
  if (s_mod((float)num, mm) != 0)
    median = list[num/2];
  else
    median = (list[num/2 - 1] + list[num/2]) / 2.0;

  return(median);
}

/* smooth orig into xx, both having size nn, rebin factor 'factor'.
 */
static void l_smooth (float * xx, float * orig, int nn, int factor)
{
  float mean = 0.0;
  int   i;

  float med_list[64];
  
  int   quot;
  int   start;
  int   rem;
  int   j;
  

  if (factor == 1)
    {
      memcpy(xx, orig, nn * sizeof(float));
      return;
    }

  /* -- fast and dirty --
   */
  if (smooth_mode == S_AVERAGE)
    {
      /* -- running average -- */
      mean = 0;
      
      for (i = 0; i < factor; ++i)
	mean += orig[i];
      
      xx[0] = mean / factor;
      
      for (i = 1; i < (nn - factor); ++i)
	{
	  mean -= orig[i-1];
	  mean += orig[i-1+factor];
	  xx[i] = mean / factor;
	}
      
      for (i = (nn    - factor); i <    nn; ++i)
	xx[i] = mean / factor;
      
      /* -- end running average -- */
    }

  else if (smooth_mode == S_MEDIAN)
    {
      if (factor > 64)
	{
	  fprintf(stderr, "* Warning *: Median limited to 64 max.\n");
	  factor = 64;
	}

      memcpy (med_list, orig, factor * sizeof(float));
      mean  = s_median (med_list, factor);
      for (i = 0; i < (factor/2 + 1); ++i)
	xx[i] = mean;
      
      for (i = (factor/2 + 1); i < (nn-factor); ++i)
	{
	  memcpy (med_list, &orig[i - (factor/2)], factor * sizeof(float));
	  /* -- median -- */

	  mean  = s_median (med_list, factor);

	  xx[i] = mean;
	}
      for (i = (nn-factor); i < nn; ++i)
	xx[i] = mean;
    }

  else if (smooth_mode == S_REBIN)
    {
      quot = nn / factor;
      rem  = nn - (quot * factor);
      
      for (i = 0; i < quot; ++i)
	{
	  mean  = 0;
	  start = i * factor;
	  
	  for (j = 0; j < factor; ++j)
	    mean += orig[start+j];
	  
	  mean /= factor;
	  
	  for (j = 0; j < factor; ++j)
	    xx[start+j] = mean;
	}
      
      start = quot * factor;
      mean  = 0;
      
      for (j = 0; j < rem; ++j)
	mean += orig[start+j];
      
      mean /= factor;
      
      for (j = 0; j < rem; ++j)
	xx[start+j] = mean;
    }

  return;
}

      
/* ----------------------------------------------- */

int getlambda(float * xposition, float * yposition, 
	      int * numlines)
{
  int lines = 0;

  int i;

  float   xmax = 1e-8;
  float   xmin = 1e+8;
  float   ymax = 1e-8;
  float   ymin = 1e+8;

  float   uxmax;
  float   uxmin;
  float   uymax;
  float   uymin;

  int     haveit, index;
  float   uxpos, uypos;
  char    c;
  float * wav, * flux;
  float * l_wave, * l_flux;
  int n_count, l_count;  
  *yposition=sbg;

  index = 0;
  ymax = max(&sp); ymin = min(&sp);
  xmax = max(&wave); xmin = min(&wave);

  wav     = copy_to_float(&wave);
  flux    = copy_to_float(&sp);
  n_count = sp.len;  

  l_wave  = (float *) malloc (sizeof(float) * n_count);
  l_flux  = (float *) malloc (sizeof(float) * n_count);
  l_count = n_count;

  if (l_wave == NULL || l_flux == NULL)
    {
      fprintf(stderr, "* Warning *: Out of memory\n");
      return (-1);
    }

  memcpy (l_wave, wav,  n_count * sizeof(float));
  memcpy (l_flux, flux, n_count * sizeof(float));

  if (cpgopen("/XSERVE") != 1) 
    {
      fprintf(stderr, "* Warning *: Cannot open plot device\n");
      free(wav); free(flux);
      return (-1);
    }

  uxmax = xmax;
  uymax = ymax;
  uxmin = xmin;
  uymin = ymin;

  uxpos = (uxmax+uxmin)/2;
  uypos = (uymax+uymin)/2;

  redraw_it (xposition, *yposition, lines,
	wav, flux, n_count,
	uxmin, uxmax, uymin, uymax);

  haveit = 0;

  while (1)
    {

      cpgband (7, 1, (uxmax+uxmin)/2, (uymax+uymin)/2, &uxpos, &uypos, &c);

      if (c == 'q' || c == 'Q')
	break;

      if (c == 'c')
        {
	  zoom  = 1.0;

          uxmax = xmax;
          uymax = ymax;
          uxmin = xmin;
          uymin = ymin;

          uxpos = (uxmax+uxmin)/2;
          uypos = (uymax+uymin)/2;

          haveit = 0;

	  redraw_it (xposition, *yposition, lines,
		     l_wave, l_flux, l_count,
		     uxmin, uxmax, uymin, uymax);
          continue;
        }

      /* zoom in (>) or out (<)
       */
      if (c == '>' || c == '<')
	{
	  haveit = 0;
	  
	  if (c == '>')
	    zoom *= 2.0;
	  else 
	    zoom /= 2.0;

	  zoom = (zoom < 1.0) ? 1.0 : zoom;

	  uymax = (uxmax + uxmin)/2;
	  uymin = (xmax - xmin)  / zoom;
	  uymin = uymin / 2.0;
	  uymin = (uymin < 0.0) ? (-uymin) : uymin; 

	  uxmax = uymax + uymin;
	  uxmin = uymax - uymin;

	  uymax = 1e-8;
	  uymin = 1e+8;

	  for (i = 0; i < l_count; ++i)
	    {
	      if (wav[i] >= uxmin && wav[i] <= uxmax)
		{
		  uymin = MIN(uymin, flux[i]);
		  uymax = MAX(uymax, flux[i]);
		}
	    }
	  redraw_it (xposition, *yposition, lines,
		     l_wave, l_flux, l_count,
		     uxmin, uxmax, uymin, uymax);
	  continue;
	}
      
      /* slide down (,) or up (.)
       */
      if (c == ',' || c == '.')
	{
	  haveit = 0;
	  
	  uymax = (uxmax - uxmin);
	  
	  if (c == '.')
	      uxmax += uymax * 0.8;
	  else
	      uxmax -= uymax * 0.8;

	  uxmax = MIN(xmax,         uxmax);
	  uxmax = MAX((xmin+uymax), uxmax);

	  uxmin = uxmax - uymax;
	  
	  uymax = 1e-8;
	  uymin = 1e+8;

	  for (i = 0; i < l_count; ++i)
	    {
	      if (wav[i] >= uxmin && wav[i] <= uxmax)
		{
		  uymin = MIN(uymin, flux[i]);
		  uymax = MAX(uymax, flux[i]);
		}
	    }
	  redraw_it (xposition, *yposition, lines,
		     l_wave, l_flux, l_count,
		     uxmin, uxmax, uymin, uymax);
	  continue;
	}
      
      /* --- cycle the smoothing mode ---
       */
      if (c == 's')
	{
	  ++smooth_mode;
	  if (smooth_mode > S_REBIN)
	    smooth_mode = 0;
	  l_smooth(l_wave, wav,  l_count, smooth);
	  l_smooth(l_flux, flux, l_count, smooth);
	  redraw_it (xposition, *yposition, lines,
		     l_wave, l_flux, l_count,
		     uxmin, uxmax, uymin, uymax);
	  continue;
	}

      /* --- smooth ---
       */
      if (c == '+')
	{
	  if (smooth < l_count/2)
	    {
	      ++smooth;
	      l_smooth(l_wave, wav,  l_count, smooth);
	      l_smooth(l_flux, flux, l_count, smooth);
	      redraw_it (xposition, *yposition, lines,
			 l_wave, l_flux, l_count,
			 uxmin, uxmax, uymin, uymax);
	    }
	  continue;
	}

      /* --- unsmooth ---
       */
      if (c == '-')
	{
	  if (smooth > 1)
	    {
	      --smooth;
	      l_smooth(l_wave, wav,  l_count, smooth);
	      l_smooth(l_flux, flux, l_count, smooth);
	      redraw_it (xposition, *yposition, lines,
			 l_wave, l_flux, l_count,
			 uxmin, uxmax, uymin, uymax);
	    }
	  continue;
	}

      if (c == ' ')
        {
	  fprintf(stderr, "X = %f, Y = %f\n", uxpos, uypos);
	  continue;
         }

      if (c == MOUSE_M || c == 'b')
	{
	  *yposition=uypos;
	  if (verbose) fprintf(stderr, "Background = %f\n",*yposition);
	  redraw_it (xposition, *yposition, lines,
		     l_wave, l_flux, l_count,
		     uxmin, uxmax, uymin, uymax);
	  continue;
	}

      if (c == 'd')
	{
	  if (verbose) fprintf(stderr, "%02d  Lambda = %f deleted\n", 
		  lines, xposition[lines-1]);
	  if (lines > 0) --lines;
	  redraw_it (xposition, *yposition, lines,
		     l_wave, l_flux, l_count,
		     uxmin, uxmax, uymin, uymax);
	  continue;
	  
	}

      if (c == MOUSE_R || c == 'm')
	{
	  xposition[lines]=uxpos;
	  if (verbose)
             fprintf(stderr, "%02d  Lambda = %f\n", lines+1, xposition[lines]);
	  ++lines;
	  if (lines == MAXLINES)
	   {
	    printf("* Warning *: program cannot handle more than %d lines\n",MAXLINES);
	    printf("          The maximum number of lines can be changed in cora_fit.h\n");
	    printf("          You need to change the line\n");
	    printf("          #define MAXLINES %d\n",MAXLINES);
	    printf("          After doing that you need to type 'make clean' before recompiling with make\n");
	    break;
	   }
	  cpgsci(3);
	  for (i = 0; i < lines; ++i)
	    {
	      cpgmove(xposition[i], uymin);
	      cpgdraw(xposition[i], uymax);
	    }
	  cpgsci(1);
	  continue;
	}
      else  if (c == MOUSE_L || c == 'a')
	++haveit;
      else
	continue;
      
      if (haveit == 1 && c != 'c')
	{
	  uxmin = MAX(xmin, uxpos);
	}
      
      if (haveit == 2 && c != 'c')
	{
	  haveit = 0;
	  
	  uxmax = MIN(xmax, uxpos);
	  
	  if (uxmax < uxmin)
	    {
	      uymin = uxmin; uxmin = uxmax; uxmax = uymin;
	    }

	  uymax = 1e-8;
	  uymin = 1e+8;

	  for (i = 0; i < l_count; ++i)
	    {
	      if (wav[i] >= uxmin && wav[i] <= uxmax)
		{
		  uymin = MIN(uymin, flux[i]);
		  uymax = MAX(uymax, flux[i]);
		}
	    }
	  redraw_it (xposition, *yposition, lines,
		     l_wave, l_flux, l_count,
		     uxmin, uxmax, uymin, uymax);
	}
      
    }
  cpgend();
  
  *numlines = lines; 
 
  if (*numlines<1)
     printf("* Warning *: %d Lines selected\n", *numlines);

  free(wav); free(flux);
  return(0);
}

/* ----------------------------------------------- */

void 
graph ()
{
  float xpos[MAXLINES];
  float ypos = 0.0,b_test=0.0;
  int i = 0, graph_lines, n_count;

  n_count = sp.len;

  if (0 == getlambda (xpos, &ypos, &graph_lines))
   {
    if (verbose>1)
      printf("Message      : xpos(0)=%f ;  %d lines selected\n", xpos[0], graph_lines);
   }
  else
    {
      printf("* Warning *: Datafile %s missing.\n", datafile);
      return;
    }

  if(graph_lines>0)
   {
    b_test = max(&lsigma);

    if (verbose > 1)
      printf("Message      : reallocating vectors dependent on number of lines\n");
    vec_free (&llambda);
    vec_free (&erlam);
    vec_free (&lsigma);
    vec_free (&ersig);
    vec_free (&ampout);
    vec_free (&eramp);
    vec_free (&steplam);
    vec_free (&stepsig);

    vec_init (&llambda);
    vec_init (&erlam);
    vec_init (&lsigma);
    vec_init (&ersig);
    vec_init (&ampout);
    vec_init (&eramp);
    vec_init (&steplam);
    vec_init (&stepsig);

    nlines = graph_lines;

    vec_alloc (&llambda, nlines);
    vec_alloc (&erlam, nlines);
    vec_alloc (&lsigma, nlines);
    vec_alloc (&ersig, nlines);
    vec_alloc (&ampout, nlines);
    vec_alloc (&eramp, nlines);
    vec_alloc (&steplam, nlines-1);
    vec_alloc (&stepsig, nlines-1);

    for (i = 0; i < nlines; ++i)
	{
	  ersig.vec[i]=-1.;
	  erlam.vec[i]=-1.;
	  eramp.vec[i]=-1.;
	  llambda.vec[i] = xpos[i];
	  lsigma.vec[i]  = b_test; 
	}
   }
  else
   {
    fprintf(stderr,"* Warning *: no lines entered\n");
    fprintf(stderr,"           : reading lines from input file\n");
    nlines=0;
    return;
   }

  if (ypos != 2.*(wave.vec[n_count-1]-wave.vec[0])/n_count && ypos != sbg)
   {
    i=0;
    while(fabs(wave.vec[i]-llambda.vec[0])>0.1)
     {
        b_test=(bkgd.vec[i]+bkgd.vec[i+1]+bkgd.vec[i+3])/3.;
        ++i;
     }

    if (ypos>=b_test)
       {
	sbg = ypos-b_test;
	sbg = sbg/(wave.vec[wave.len-1]-wave.vec[0])*wave.len;
       }

    if (verbose) fprintf(stderr, "sbg = %f cts/A\n",sbg);
   }
return;
}

